SystemOrganization addCategory: #'Tweak-Compiler'!

Exception subclass: #CDeprecatedFeature
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Compiler'!

!CDeprecatedFeature methodsFor: 'as yet unclassified' stamp: 'ar 8/10/2003 20:31'!
defaultAction
	"By default, allow using deprecated features"
	^true! !

!CDeprecatedFeature methodsFor: 'as yet unclassified' stamp: 'ar 8/10/2003 20:34'!
isResumable
	^true! !

Compiler subclass: #CCompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Compiler'!

!CCompiler class methodsFor: 'evaluating' stamp: 'ar 3/27/2004 04:18'!
evaluate: textOrString notifying: aController logged: logFlag in: envt
	| val |
	val := self new
				evaluate: textOrString
				notifying: aController
				in: envt.
	logFlag ifTrue: [Smalltalk logChange: textOrString].
	^val! !

!CCompiler class methodsFor: 'accessing' stamp: 'eem 5/13/2008 11:38'!
parserClass
	"Answer a parser class to use for parsing methods compiled by instances of the receiver."

	^CParser! !

!CCompiler methodsFor: 'public access' stamp: 'ar 3/18/2005 20:40'!
evaluate: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: failBlock
	^self evaluate: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: failBlock logged: true! !

!CCompiler methodsFor: 'public access' stamp: 'sn 5/25/2013 00:44'!
evaluate: textOrString notifying: aRequestor in: scope
	| methodNode method |
	"class := scope."
	self from: textOrString class: scope context: nil notifying: aRequestor.
	methodNode := self translate: sourceStream noPattern: true ifFail:[^nil].
	method := methodNode generate: #(0 0 0 0).
	self interactive ifTrue:
		[method := method copyWithTempsFromMethodNode: methodNode].
	^scope uniqueInstance withArgs: #() executeMethod: method! !

!Metaclass methodsFor: '*Tweak-Compiler' stamp: 'ar 3/27/2004 06:06'!
compilerClass
	^thisClass compilerClass! !

!Metaclass methodsFor: '*Tweak-Compiler' stamp: 'ar 3/27/2004 06:07'!
evaluatorClass
	^thisClass evaluatorClass! !

!Metaclass methodsFor: '*Tweak-Compiler' stamp: 'ar 3/27/2004 06:06'!
parserClass
	^thisClass parserClass! !

Parser subclass: #CScanner
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Compiler'!

!CScanner commentStamp: '<historical>' prior: 0!
I am a subclass of Scanner for tokenising for the Tweak compiler.  Since I need to be subclassed by CParser, which also needs to subclass Parser, and since we don't have multiple inheritance I directly inherit from Parser, whose superclass is Scanner.!

CScanner subclass: #CParser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Compiler'!

!CParser methodsFor: 'private' stamp: 'eem 12/4/2008 18:29'!
addProperty: propName value: value
	| aSymbol |
	aSymbol := propName asSymbol.
	(properties includesKey: aSymbol) ifTrue:[self notify: '<- duplicate property'].
	properties := properties copyWith: (Association key: aSymbol value: value).
	^true! !

!CParser methodsFor: 'error handling' stamp: 'eem 8/30/2010 14:19'!
addWarning: aString
	properties := self properties copyWith: (Association key: #warning value: aString)! !

!CParser methodsFor: 'events' stamp: 'ar 3/26/2004 15:45'!
blockTrigger
	"Parse a block trigger"
	| methodNode method |
	self advance.
	self statements: #() innerBlock: false.
	(self match: #rightBracket) ifFalse: [^self expected: 'Period or right bracket'].
	methodNode := self newMethodNode.
	methodNode
		selector: #blockTrigger
		arguments: #()
		precedence: 1
		temporaries: #()
		block: parseNode
		encoder: encoder
		primitive: 0
		properties: nil.
	method := methodNode generate: #(0 0 0 0).
	^method
! !

!CParser methodsFor: 'error correction' stamp: 'ar 12/27/2004 11:05'!
correctSelector: proposedKeyword wordIntervals: spots exprInterval: expInt ifAbort: abortAction
	"Correct the proposedKeyword to some selector symbol, correcting the original text if such action is indicated.  abortAction is invoked if the proposedKeyword couldn't be converted into a valid selector.  Spots is an ordered collection of intervals within the test stream of the for each of the keyword parts."

	| alternatives aStream choice correctSelector userSelection lines firstLine |
	"If we can't ask the user, assume that the keyword will be defined later"
	self interactive ifFalse: [ ^ proposedKeyword asSymbol ].

	userSelection := requestor selectionInterval.
	requestor selectFrom: spots first first to: spots last last.
	requestor select.
	alternatives := Symbol possibleSelectorsFor: proposedKeyword.

	aStream := WriteStream on: Array new.
	aStream nextPut: (proposedKeyword contractTo: 35).
	firstLine := 1.
 	alternatives do:
		[:sel | aStream nextPut: (sel contractTo: 35)].
	aStream nextPut: 'cancel'.
	lines := Array with: firstLine with: (alternatives size + firstLine).
	
	choice := UIManager default chooseFrom: aStream contents lines: lines title:
'Unknown selector, please 
confirm, correct, or cancel'.

	(choice = 0) | (choice > (lines at: 2))
		ifTrue: [ ^ abortAction value ].

	requestor deselect.
	requestor selectInvisiblyFrom: userSelection first to: userSelection last.

	choice = 1 ifTrue: [ ^ proposedKeyword asSymbol ].
	correctSelector := alternatives at: choice - 1.
	self substituteSelector: correctSelector keywords wordIntervals: spots.
	((proposedKeyword last ~~ $:) and: [correctSelector last == $:]) ifTrue: [
		^ abortAction value].
	^ correctSelector.
! !

!CParser methodsFor: 'events' stamp: 'ar 3/26/2004 15:45'!
eventSignaler
	"Process an event trigger pragma, e.g.,
		<signals: #eventName> (deprecated form)

		<signals: eventName> (new form)
		<signals: eventName in: signaler>
	"
	"All style event trigger"
	| eventName var |
	hereType == #literal ifTrue:[
		"Deprecated"
		(CDeprecatedFeature signal == true) ifFalse:[^false].
		^self addProperty: #signals value: self advance].
	hereType == #word ifFalse:[^false].
	eventName := self advance.
	(hereType == #keyword and:[here = 'in:']) 
		ifFalse:[^self addProperty: #signals value: eventName asSymbol].
	self advance.
	hereType == #word ifFalse:[^self expected: 'signaler'].
	var := self variable.
	^self addProperty: #signals:in: value: {eventName asSymbol. var name asSymbol}! !

!CParser methodsFor: 'events' stamp: 'ar 3/26/2004 15:45'!
eventTrigger
	"Process an event trigger pragma, e.g.,
		<on: #eventName> (deprecated form)

		<on: eventName> (new form)
		<on: eventName in: signaler>
	"
	"All style event trigger"
	| eventName var |
	hereType == #literal ifTrue:[
		"Deprecated"
		(CDeprecatedFeature signal == true) ifFalse:[^false].
		^self addProperty: 'on' value: self advance].
	hereType == #leftBracket ifTrue:[^self addProperty: 'on' value: self blockTrigger].
	hereType == #word ifFalse:[^false].
	eventName := self advance.
	(hereType == #keyword and:[here = 'in:']) 
		ifFalse:[^self addProperty: #on value: eventName asSymbol].
	self advance.
	hereType == #word ifFalse:[^self expected: 'signaler'].
	var := self variable.
	^self addProperty: #on:in: value: {eventName asSymbol. var name asSymbol}! !

!CParser methodsFor: 'expression types' stamp: 'ar 4/28/2004 16:39'!
expression

	(hereType == #word and: [tokenType == #leftArrow])
		ifTrue: [^ self assignment: self variable].
	(hereType == #word and: [tokenType == #leftBracket])
		ifTrue:[^self matrixExpression: false].
	hereType == #leftBrace
		ifTrue: [self braceExpression]
		ifFalse: [self primaryExpression ifFalse: [^ false]].
	(self messagePart: 3 repeat: true)
		ifTrue: [hereType == #semicolon ifTrue: [self cascade]].
	^self remoteAssignment
! !

!CParser methodsFor: 'properties' stamp: 'ar 3/26/2004 15:45'!
hasProperty: propName
	properties ifNil:[^false].
	^properties includesKey: propName! !

!CParser methodsFor: 'expression types' stamp: 'eem 8/31/2010 17:42'!
method: doit context: ctxt
	" pattern [ | temporaries ] block => MethodNode."

	| sap blk prim temps messageComment methodNode |
	sap := self pattern: doit inContext: ctxt.
	"sap={selector, arguments, precedence}"
	self properties selector: (sap at: 1).
	encoder selector: (sap at: 1).
	(sap at: 2) do: [:argNode | argNode beMethodArg].
	temps := self temporaries.
	messageComment := currentComment.
	currentComment := nil.
	doit ifFalse: [self tweakProperties].
	prim := self pragmaPrimitives.
	self statements: #() innerBlock: doit.
	blk := parseNode.
	doit ifTrue: [blk returnLast]
		ifFalse: [blk returnSelfIfNoOther: encoder].
	hereType == #doIt ifFalse: [^self expected: 'Nothing more'].
	methodNode := self newMethodNode comment: messageComment.
	methodNode
		selector: (sap at: 1)
		arguments: (sap at: 2)
		precedence: (sap at: 3)
		temporaries: temps
		block: blk
		encoder: encoder
		primitive: prim
		properties: properties.
	self interactive ifTrue:
		[self declareUndeclaredTemps: methodNode.
		 self removeUnusedTemps].
	^methodNode! !

!CParser methodsFor: 'events' stamp: 'ar 3/26/2004 15:45'!
parseArgNames: aString
        "Parse the argument, aString, notifying req if an error occurs. Otherwise, 
        answer a two-element Array containing Arrays of strings (the argument 
        names and temporary variable names)."
        aString == nil ifTrue: [^#()].
        doitFlag := false.               "Don't really know if a doit or not!!"
        ^self initPattern: aString
                notifying: nil
                return: [:pattern | (pattern at: 2)]! !

!CParser methodsFor: 'public access' stamp: 'ar 3/26/2004 15:45'!
parseArgs: aString notifying: req 
 	"Parse the argument, aString, notifying req if an error occurs. Otherwise, 
 	answer an Array containing the argument names"

	aString == nil ifTrue: [^#()].
	doitFlag := false.               "Don't really know if a doit or not!!"
	^self initPattern: aString
		notifying: req
		return: [:pattern | (pattern at: 2)]! !

!CParser methodsFor: 'primitives' stamp: 'eem 8/30/2010 14:20'!
primitiveDeclaration
	| prim module |
	(self matchToken: 'primitive:') ifFalse:[^false].
	prim := here.
	(self match: #number) ifTrue:[ "Indexed primitives"
		^self addPragma: (Pragma keyword: #primitive: arguments: (Array with: prim))].
	(self match: #string) ifFalse:[^self expected:'Integer or String'].
	(self matchToken: 'module:') ifTrue:[
		module := here.
		(self match: #string) ifFalse:[^self expected: 'String'].
		module := module asSymbol].
	self allocateLiteral: (Array with: module with: prim asSymbol with: 0 with: 0).
	^self addPragma: (Pragma keyword: #primitive: arguments: (Array with: 117))! !

!CParser methodsFor: 'properties' stamp: 'eem 8/30/2010 14:07'!
property
	"Read a single property. Primitives and others may have fixed spec.
	Parse all generic properties in the form of:
		<propKey: propValue>
	and remember them as method properties."
	| propName propValue varStart varEnd |
	(self hasProperty: #primitive:) ifFalse:[
		"Only one primitive specification is allowed"
		self externalFunctionDeclaration ifTrue:[^true].
		self primitiveDeclaration ifTrue:[^true].
	].
	hereType == #keyword ifFalse:[^false].
	propName := self advance allButLast. "remove last colon"
	propName = 'on' ifTrue:[^self eventTrigger].
	"propName = 'signals' ifTrue:[^self eventSignaler]."
	(hereType == #number or:[hereType == #literal or:[hereType == #string]]) ifTrue:[
		^self addProperty: propName value: self advance].
	"We shouldn't have any but #word hereTypes at this point"
	hereType == #word ifFalse:[^self error:'*** FIX THIS ***'].

	varStart := self startOfNextToken + requestorOffset.
	propValue := self advance.
	varEnd := self endOfLastToken + requestorOffset.

	propValue = 'true' ifTrue:[^self addProperty: propName value: true].
	propValue = 'false' ifTrue:[^self addProperty: propName value: false].

	[true] whileTrue:[
		Symbol hasInterned: propValue ifTrue:[:aSymbol|
			(encoder classEncoding bindingOf: aSymbol) ifNotNilDo:[:assoc|
				^self addProperty: propName value: assoc value.
			].
		].
		propValue := self correctProperty: propValue interval: (varStart to: varEnd).
		propValue ifNil:[^nil].
	].
	^true! !

!CParser methodsFor: 'expression types' stamp: 'eem 5/12/2008 13:50'!
remoteAssignment
	| leftHandNode selector start |
	hereType == #leftArrow ifFalse:[^true].

	parseNode isMessageNode ifFalse:[^true].
	parseNode arguments size = 0 ifFalse:[^true].
	leftHandNode := parseNode.
	selector := leftHandNode selector key, ':'.
	(Symbol hasInterned: selector ifTrue: [ :sym | selector := sym]) ifFalse:[
		"Argh... how would we correct this???"
		selector := selector asSymbol.
	].
	self advance.
	self expression ifFalse:[^false].
	start := (encoder sourceRangeFor: leftHandNode) first.
	parseNode := MessageNode new
				receiver: leftHandNode receiver
				selector: selector
				arguments: (Array with: parseNode)
				precedence: 3
				from: encoder
				sourceRange: (start to: self endOfLastToken).
	^true! !

!CParser methodsFor: 'expression types' stamp: 'ar 3/27/2004 14:03'!
scopeVariable

	| varName varStart varEnd scope binding |
	varStart := self startOfNextToken + requestorOffset.
	scope := encoder classEncoding.
	[varName := self advance.
	binding := scope bindingOf: varName asSymbol.
	binding == nil ifTrue:[^self expected: 'variable name'].
	hereType == #scopeRef] whileTrue:[
		self advance. "skip scopeRef"
		scope := binding value].
	varEnd := self endOfLastToken + requestorOffset.
	^encoder encodeGlobal: binding sourceRange: (varStart to: varEnd)
! !

!CParser methodsFor: 'properties' stamp: 'ar 2/12/2010 18:54'!
tweakProperties
	"Parse method properties"
	[true] whileTrue:[
		(self matchToken: #<) ifFalse:[^self].
		self property.
		(self matchToken: #>) ifFalse: [^ self expected: '>'].
	].! !

!CParser methodsFor: 'expression types' stamp: 'eem 5/13/2008 14:39'!
variable

	^tokenType == #scopeRef
		ifTrue:[self scopeVariable]
		ifFalse: [super variable]! !

!CScanner class methodsFor: 'testing' stamp: 'eem 5/13/2008 15:59'!
isLiteralSymbol: aSymbol 
	"Test whether a symbol can be stored as # followed by its characters.  
	Symbols created internally with asSymbol may not have this property, 
	e.g. '3' asSymbol."
	| typeTable i ascii type |
	typeTable := superclass classPool at: #TypeTable. "yuck!!"
	i := aSymbol size.
	i = 0 ifTrue: [^ false].
	i = 1 ifTrue: [('$''"()#0123456789' includes: (aSymbol at: 1))
		ifTrue: [^ false] ifFalse: [^ true]].
	ascii := (aSymbol at: 1) asciiValue.
	"TypeTable should have been origined at 0 rather than 1 ..."
	ascii = 0 ifTrue: [^ false].
	type := typeTable at: ascii.
	(type == #xColon or: [type == #verticalBar]) ifTrue: [^ i = 1].
	type == #xBinary ifTrue: 
			[[i > 1]
				whileTrue: 
					[ascii := (aSymbol at: i) asciiValue.
					ascii = 0 ifTrue: [^ false].
					(typeTable at: ascii) == #xBinary ifFalse: [^ false].
					i := i - 1].
			^ true].
	type == #xLetter ifTrue: 
			[[i > 1]
				whileTrue: 
					[ascii := (aSymbol at: i) asciiValue.
					ascii = 0 ifTrue: [^ false].
					type := typeTable at: ascii.
					(type == #xLetter or: [type == #xDigit or: [type == #xColon]])
						ifFalse: [^ false].
					i := i - 1].
			^ true].
	^ false! !

!CScanner methodsFor: 'public access' stamp: 'ar 3/26/2004 15:45'!
scanTokens: textOrString 
	"Answer an Array that has been tokenized as though the input text, 
	textOrString, had appeared between the array delimitors #( and ) in a 
	Smalltalk literal expression."

	self scan: (ReadStream on: textOrString asString).
	self scanLitVec.
	^token

	"Scanner new scanTokens: 'identifier keyword: 8r31 ''string'' .'"! !

!CScanner methodsFor: 'multi-character scans' stamp: 'ar 5/10/2005 12:24'!
xBinary

	tokenType := #binary.
	token := self step asSymbol.
	[(self typeTableAt: hereChar) = #xBinary and: [hereChar ~= $-]]
		whileTrue: [token := (token , (String with: self step)) asSymbol]! !

!CScanner methodsFor: 'multi-character scans' stamp: 'eem 5/12/2008 11:18'!
xColon		"Allow := for assignment by converting to #_ "
	aheadChar = $= ifTrue:
		[self step.
		tokenType := #leftArrow.
		self step.
		^ token := #':='].
	aheadChar = $: ifTrue:
		[self step.
		tokenType := #scopeRef.
		self step.
		^ token := #'::'].		
	"Otherwise, just do what normal scan of colon would do"
	tokenType := #colon.
	^ token := self step asSymbol! !

!CScanner methodsFor: 'multi-character scans' stamp: 'ar 5/10/2005 12:23'!
xLetter
	"Form a word or keyword."

	| type |
	buffer reset.
	[(type := self typeTableAt: hereChar) == #xLetter or: [type == #xDigit]]
		whileTrue:
			["open code step for speed"
			buffer nextPut: hereChar.
			hereChar := aheadChar.
			source atEnd
				ifTrue: [aheadChar := 30 asCharacter "doit"]
				ifFalse: [aheadChar := source next]].
	(type == #colon or: [type == #xColon and: [aheadChar ~= $= and:[aheadChar ~= $:]]])
		ifTrue: 
			[buffer nextPut: self step.
			["Allow any number of embedded colons in literal symbols"
			(self typeTableAt: hereChar) == #xColon]
				whileTrue: [buffer nextPut: self step].
			tokenType := #keyword]
		ifFalse: 
			[type == #leftParenthesis 
				ifTrue:[buffer nextPut: self step; nextPut: $).
						tokenType := #positionalMessage]
				ifFalse:[tokenType := #word]].
	token := buffer contents! !

!Class methodsFor: '*Tweak-Compiler' stamp: 'sn 10/2/2011 13:42'!
fullName
	^(environment == nil or:[environment == Smalltalk])
		ifTrue:[^('Squeak::',name) asSymbol]
		ifFalse:[
			(environment class = SystemDictionary) ifTrue:[^self name] ifFalse:[
			
			^environment fullNameFor: self localName]].
! !

!Class methodsFor: '*Tweak-Compiler' stamp: 'ar 3/27/2004 14:06'!
localName
	^(name findTokens:':') last asSymbol! !
